# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2017-2025
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This plugin implements the ``kas lock`` command.

    When this command is executed a locking spec is created which only contains
    the exact commit of each repository. This is used to pin the commit of
    floating branches and tags, while still keeping an easy update path. The
    lockfile is created next to the first file on the kas cmdline. For details
    on the locking support, see :class:`kas.includehandler.IncludeHandler`.

    .. note ::
       * all referenced repositories are checked out to resolve cross-repo
         configs
       * all branches are resolved before patches are applied

    **Updating lockfiles**

    When updating lockfiles, kas attempts to update the repository revisions
    in the lockfile that defines the revision. If a repository is exclusively
    locked in an external lockfile, this lock is not updated (we cannot modify
    an external repository). However, if the revision is also defined in a
    local lockfile, it is updated in the local lockfile.

    The algorithm for determining where to pin the revision of a repository
    is as follows:

    #. Find all repositories that have a floating ref (i.e. no commit). Assign
       to `to-lock` list
    #. Iterate over all lockfiles (in include order)

       #. for each repository, check if it is locked in the current file

          #. if lock is up to date, remove from `to-lock` list
          #. else if lockfile is internal, update lockfile, remove from
             `to-lock` list
          #. else (repo is locked in external lockfile): mark repo external

    #. Remove all repos with `external` marks from `to-lock` list
    #. Add all remaining repos in `to-lock` list to topmost lockfile,
       create if needed

    **Examples**

    The lockfile is created as ``kas-project.lock.yml``.
    Call again to regenerate lockfile::

        kas lock --update kas-project.yml

    The generated lockfile will automatically be used to pin the revisions::

        kas build kas-project.yml

    Note, that the lockfiles should be checked-in into the VCS.
"""

import logging
import os
from dataclasses import dataclass
from kas.context import get_context
from kas.includehandler import ConfigFile
from kas.plugins.checkout import Checkout
from kas.plugins.dump import Dump, IoTarget, LOCKFILE_VERSION_MIN
from kas.plugins.diff import Diff
from kas.repos import Repo, RepoRefError

__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens AG, 2024'


class Lock(Checkout):
    """
    Implements a kas plugin to create and update kas project lockfiles.
    """

    name = 'lock'
    helpmsg = (
        'Create and update kas project lockfiles.'
    )

    @dataclass
    class RepoInfo():
        repo: Repo
        ext_lock: bool

    @classmethod
    def setup_parser(cls, parser):
        super().setup_parser(parser)
        Dump.setup_parser_format_args(parser)

    def _print_log_diff(self, repo, old_commit):
        try:
            diff = repo.diff(old_commit, None)
        except NotImplementedError:
            return
        except RepoRefError as e:
            logging.warning(e)
            return
        Diff.formatting_diff_output(
            None, None, {'vcs': diff}, True, False, True, False)

    def _update_lockfile(self, lockfile, repos_to_lock, update_only, args):
        """
        Update all locks in the given lockfile.
        If update_only, no new locks are added.
        """
        output = IoTarget(target=lockfile, managed=True)
        lockfile_config = lockfile.config
        changed = False

        if 'overrides' not in lockfile_config:
            lockfile_config['overrides'] = {'repos': {}}
        if 'repos' not in lockfile_config['overrides']:
            lockfile_config['overrides']['repos'] = {}
        lock_header_vers = lockfile_config['header']['version']
        if lock_header_vers < LOCKFILE_VERSION_MIN:
            logging.warning('Lockfile uses too-old header version (%s). '
                            'Updating to version %d',
                            lock_header_vers, LOCKFILE_VERSION_MIN)
            lockfile_config['header']['version'] = LOCKFILE_VERSION_MIN

        for k, v in lockfile_config['overrides']['repos'].items():
            ri = repos_to_lock.get(k)
            if not ri:
                continue

            r = ri.repo
            if v['commit'] == r.revision or \
                    v['commit'] == r.revision_annotated_tag:
                logging.info('Lock of %s is up-to-date: %s',
                             r.name, v['commit'])
            elif not lockfile.is_external:
                logging.info('Updating lock of %s: %s -> %s',
                             r.name, v['commit'], r.revision)
                self._print_log_diff(r, v['commit'])
                v['commit'] = r.revision
                changed = True
            else:
                logging.warning(
                    'Repo %s is locked in remote lockfile %s. '
                    'Not updating.', r.name, lockfile.filename)
                ri.ext_lock = True
                continue
            del repos_to_lock[k]

        if not update_only:
            for k, ri in repos_to_lock.items():
                r = ri.repo
                logging.info('Adding lock of %s: %s', r.name, r.revision)
                lockfile_config['overrides']['repos'][k] = \
                    {'commit': r.revision}
                changed = True

        if not changed:
            return repos_to_lock

        logging.info('Updating lockfile %s',
                     os.path.relpath(lockfile.filename, os.getcwd()))
        output = IoTarget(target=lockfile.filename, managed=True)
        format = "json" if lockfile.filename.suffix == '.json' else "yaml"
        Dump.dump_config(lockfile_config, output, format,
                         args.indent, args.sort)
        return repos_to_lock

    def run(self, args):
        def _filter_enabled(repos):
            return [(k, r) for k, r in repos if not r.operations_disabled]

        args.skip += [
            'setup_dir',
            'repos_apply_patches',
            'setup_environ',
            'write_bbconfig',
        ]

        super().run(args)
        ctx = get_context()
        repos_cfg = ctx.config.repo_dict.items()
        # when locking, only consider floating repos managed by kas
        # Important: process repos in the order they are defined in the
        # config file to update lockfile with highest precedence.
        repos_to_lock = dict([(k, self.RepoInfo(r, False)) for k, r
                              in _filter_enabled(repos_cfg)
                              if not r.commit])
        if not repos_to_lock:
            logging.info('No floating repos found. Nothing to lock.')
            return

        # first update all locks we have without creating new ones
        lockfiles = ctx.config.get_lockfiles()
        for lock in lockfiles:
            repos_to_lock = self._update_lockfile(lock, repos_to_lock,
                                                  True, args)

        # remove repos that are externally locked
        repos_to_lock = {k: v for k, v in repos_to_lock.items()
                         if not v.ext_lock}

        # then add new locks for the remaining repos to the default lockfile
        if repos_to_lock:
            repo_to_lock_names = [r.repo.name for r in repos_to_lock.values()]
            logging.warning('The following repos are not covered by any '
                            'lockfile. Adding to top lockfile: %s',
                            ', '.join(repo_to_lock_names))
            lockpath = ctx.config.handler.get_lock_filename()
            if lockfiles and lockfiles[0].filename == lockpath:
                lock = lockfiles[0]
            else:
                lock = ConfigFile(lockpath, is_external=False,
                                  is_lockfile=True)
                lock.config['header'] = {'version': LOCKFILE_VERSION_MIN}
            self._update_lockfile(lock, repos_to_lock, False, args)


__KAS_PLUGINS__ = [Lock]



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2017-2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This plugin implements the ``kas dump`` command.

    When this command is executed in default mode, kas will parse all
    referenced config files, expand includes and print a flattened yaml version
    of the configuration to stdout. This config is semantically identical to
    the input, but does not include any references to other configuration
    files. The output of this command can be used to further analyze the build
    configuration.

    When running with ``--lock``, a locking spec is created which only contains
    the exact commit of each repository. This can be used to pin the commit of
    floating branches and tags, while still keeping an easy update path. For
    details on the locking support, see :class:`kas.plugins.lock`.

    .. note::
        The options to create and update lock files have been moved to the lock
        plugin.

    When running with ``--resolve-local``, VCS tracking information of the root
    repo (the one with the kas-project.yml) is added to the output. The
    generated file can be used as single input to kas to reproduce the build
    environment. If the root repo is not under version control or contains
    uncommitted changes, a warning is emitted.

    Please note:

    - the dumped config is semantically identical but not bit-by-bit identical
    - all referenced repositories are checked out to resolve cross-repo configs
    - all branches are resolved before patches are applied
    - the ordering of the keys is kept unless ``--sort`` is used. If you intend
      to store the flattened configs for comparison, it is recommended to sort
      the keys.

    For example, to get a single config representing the final build config of
    ``kas-project.yml:target-override.yml`` you could run::

        kas dump kas-project.yml:target-override.yml > kas-project-expanded.yml

    The generated config can be used as input for kas::

        kas build kas-project-expanded.yml
"""

import sys
import json
import yaml
import logging
import argparse
from typing import TypeVar, TextIO
from collections import OrderedDict
from kas.context import get_context
from kas.plugins.checkout import Checkout
from kas.kasusererror import KasUserError, ArgsCombinationError

__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens AG, 2022'

LOCKFILE_VERSION_MIN = 14
SCHEMA_VERSION_MIN = 7


class OutputFormatError(KasUserError):
    def __init__(self, format):
        super().__init__(f'invalid format {format}')


class IoTarget:
    StrOrTextIO = TypeVar('StrOrTextIO', str, TextIO)

    target: StrOrTextIO
    managed: bool

    def __init__(self, target, managed):
        self.target = target
        self.managed = managed


class IoTargetMonitor:
    """
    Simple monitor to unify access to file targets that need
    to be closed (files) and ambient ones (stdout / stderr)
    """

    def __init__(self, target: IoTarget):
        self._target = target
        self._file = None

    def __enter__(self):
        if self._target.managed:
            self._file = open(self._target.target, 'w')
            return self._file
        return self._target.target

    def __exit__(self, exc_type, exc_value, traceback):
        if self._target.managed:
            self._file.close()


class Dump(Checkout):
    """
    Implements a kas plugin that combines multiple kas configurations
    and dumps the result.
    """

    name = 'dump'
    helpmsg = (
        'Expand and dump the final config to stdout. When resolving branches, '
        'this is done before patches are applied.'
    )

    class KasYamlDumper(yaml.Dumper):
        """
        Yaml formatter (dumper) that generates output in a formatting which
        is similar to kas example input files.
        """

        def represent_data(self, data):
            if isinstance(data, str):
                if data.count('\n') > 0:
                    return self.represent_scalar(
                        'tag:yaml.org,2002:str',
                        data,
                        style='|')
                return self.represent_scalar('tag:yaml.org,2002:str', data)
            elif isinstance(data, OrderedDict):
                return self.represent_mapping(
                    'tag:yaml.org,2002:map',
                    data.items())
            elif data is None:
                return self.represent_scalar('tag:yaml.org,2002:null', '')
            return super().represent_data(data)

    @staticmethod
    def setup_parser_format_args(parser):
        parser.add_argument('--indent',
                            type=int,
                            default=4,
                            help='Line indent (# of spaces, default: 4)')
        parser.add_argument('--sort',
                            action='store_true',
                            default=False,
                            help='Alphanumerically sort keys in output')

    @classmethod
    def setup_parser(cls, parser):
        super().setup_parser(parser)
        Dump.setup_parser_format_args(parser)
        lk_or_env = parser.add_mutually_exclusive_group()
        parser.add_argument('--format',
                            choices=['yaml', 'json'],
                            default='yaml',
                            help='Output format (default: yaml)')
        parser.add_argument('--resolve-refs',
                            action='store_true',
                            help='Replace floating refs with exact SHAs. '
                                 'Overrides are removed')
        parser.add_argument('--resolve-local',
                            action='store_true',
                            help='Add tracking information of root repo')
        lk_or_env.add_argument('--resolve-env',
                               action='store_true',
                               help='Set env defaults to captured env value')
        lk_or_env.add_argument('--lock',
                               action='store_true',
                               help='Create lockfile with exact SHAs')
        parser.add_argument('-i', '--inplace',
                            action='store_true',
                            help=argparse.SUPPRESS)

    @staticmethod
    def dump_config(config: dict, target: IoTarget, format: str, indent: int,
                    sorted: bool):
        """
        Dump the configuration to the target in the specified format.
        """
        with IoTargetMonitor(target) as f:
            if format == 'json':
                json.dump(config, f, indent=indent, sort_keys=sorted)
                f.write('\n')
            elif format == 'yaml':
                yaml.dump(
                    config, f,
                    indent=indent,
                    sort_keys=sorted,
                    Dumper=Dump.KasYamlDumper)
            else:
                raise OutputFormatError(format)

    def run(self, args):
        def _filter_enabled(repos):
            return [(k, r) for k, r in repos if not r.operations_disabled]

        def _filter_local(repos):
            return [(k, r) for k, r in repos
                    if r.operations_disabled and r.name]

        args.skip += [
            'setup_dir',
            'repos_apply_patches',
            'setup_environ',
            'write_bbconfig',
        ]

        super().run(args)
        ctx = get_context()
        schema_v = LOCKFILE_VERSION_MIN if args.lock else SCHEMA_VERSION_MIN
        config_expanded = {'header': {'version': schema_v}} if args.lock \
            else ctx.config.get_config(remove_includes=True)
        repos = ctx.config.repo_dict.items()
        output = IoTarget(target=sys.stdout, managed=False)

        if args.inplace and not args.lock:
            raise ArgsCombinationError('--inplace requires --lock')
        if args.resolve_local and args.lock:
            raise ArgsCombinationError(
                '--resolve-local cannot be used with --lock')
        if args.inplace and args.lock:
            from kas.plugins.lock import Lock
            logging.warning('The --inplace option is deprecated. '
                            'Migrate to the "lock" command.')
            return Lock().run(args)

        if args.lock:
            args.resolve_refs = True
            # when locking, only consider repos managed by kas
            repos = _filter_enabled(repos)
            config_expanded['overrides'] = \
                {'repos': {k: {'commit': r.revision} for k, r in repos}}

        if args.resolve_refs and not args.lock:
            for k, r in _filter_enabled(repos):
                if r.commit or r.branch or r.tag:
                    config_expanded['repos'][k]['commit'] = r.revision
                elif r.refspec:
                    config_expanded['repos'][k]['refspec'] = r.revision
            # as the refs are resolved, the overrides are redundant
            if 'overrides' in config_expanded:
                del config_expanded['overrides']

        if args.resolve_local:
            for k, r in _filter_local(repos):
                if r.revision:
                    if r.dirty:
                        logging.warning(f'Repository {r.name} (root repo) '
                                        'contains uncommitted changes.')
                    if config_expanded['repos'][k] is None:
                        config_expanded['repos'][k] = {}
                    config_expanded['repos'][k]['url'] = r.url
                    config_expanded['repos'][k]['commit'] = r.revision
                else:
                    logging.warning(f'Repository {r.name} (root repo) '
                                    'is not under version control.')

        if args.resolve_env and 'env' in config_expanded:
            config_expanded['env'] = ctx.config.get_environment()

        self.dump_config(config_expanded, output, args.format, args.indent,
                         args.sort)


__KAS_PLUGINS__ = [Dump]



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens, 2025
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This plugin implements the ``kas diff`` command.

    This plugin compares two KAS configurations and outputs the
    differences. The diff includes both content differences in
    the configuration files and repository differences if commit
    IDs or tags have changed.

    Additionally, you can use the ``--format json`` option to output
    the diff in JSON format.

    .. note::
      The text output of the plugin should not considered to be stable. If
      stable output is needed, use a machine readable format like json.
"""

import json
import difflib
from kas.context import create_global_context
from kas.config import Config
from kas.libcmds import Macro
from kas.libkas import setup_parser_common_args

__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens, 2025'


class Diff:
    """
    kas plugin to compute diff of two kas configurations.
    """

    name = 'diff'
    helpmsg = (
        'Compare two KAS configurations.'
    )

    @classmethod
    def setup_parser(cls, parser):
        setup_parser_common_args(parser)

        parser.add_argument('config1',
                            help='The first config file to be compared.')
        parser.add_argument('config2',
                            help='The second config file to be compared.')
        parser.add_argument('--format',
                            choices=['json', 'text'],
                            default='text',
                            help='Diff output format (default: text)')
        parser.add_argument('--oneline',
                            action='store_true',
                            help='Use git oneline output for differing '
                                 'commits.')
        parser.add_argument('--no-color',
                            action='store_true',
                            help='Disable colored highlighting of diffs.')
        parser.add_argument('--commit-only',
                            action='store_true',
                            help='This will not display the differences in '
                                 'the KAS configurations; it will only list '
                                 'commits resulting from different '
                                 'repository revisions.')
        parser.add_argument('--content-only',
                            action='store_true',
                            help='This will only display the differences in '
                                 'the KAS configurations and will not '
                                 'include repository differences.')

    @staticmethod
    def compare_dicts(dict1, dict2, parent_key=''):
        """
        Deep compare dictionaries. Returns a dictionary with the differences.
        """
        diff = {"values_changed": {}}

        def add_change(key, old_value, new_value):
            diff["values_changed"][key] = {
                "old_value": old_value,
                "new_value": new_value
            }

        def deep_compare(d1, d2, parent_key):
            keys = set(d1.keys()).union(set(d2.keys()))
            for key in keys:
                full_key = f"{parent_key}.{key}" if parent_key else key
                if key in d1 and key in d2:
                    if isinstance(d1[key], dict) and isinstance(d2[key], dict):
                        deep_compare(d1[key], d2[key], full_key)
                    elif d1[key] != d2[key]:
                        add_change(full_key, d1[key], d2[key])
                elif key in d1:
                    add_change(full_key, d1[key], None)
                else:
                    add_change(full_key, None, d2[key])

        deep_compare(dict1, dict2, parent_key)
        return diff

    @staticmethod
    def formatting_diff_output(oldfile, newfile, diff_output, oneline,
                               no_color, commit_only, content_only):
        """
        Format the diff output.
        """
        def print_unified_diff(oldv, newv, key):
            # unified diff expects newline terminated input strings
            if '\n' not in oldv:
                oldv += '\n'
            if '\n' not in newv:
                newv += '\n'
            diff = difflib.unified_diff(oldv.splitlines(keepends=True),
                                        newv.splitlines(keepends=True))
            print(f"{' ' * 5}{key}:")
            for line in diff:
                if line.startswith('+++') or line.startswith('---'):
                    continue
                elif line.startswith('@@'):
                    print(f"{COLORS_FILES}{' ' * 9}{line}{COLORS_ENDC}",
                          end='')
                elif line[0] == '+':
                    print(f"{COLORS_NEW}+{' ' * 8}{line[1:]}{COLORS_ENDC}",
                          end='')
                elif line[0] == '-':
                    print(f"{COLORS_OLD}-{' ' * 8}{line[1:]}{COLORS_ENDC}",
                          end='')
                else:
                    print(f"{' ' * 8}{line}", end='')

        if no_color:
            COLORS_OLD = ''
            COLORS_NEW = ''
            COLORS_COMMIT = ''
            COLORS_FILES = ''
            COLORS_ENDC = ''
        else:
            COLORS_OLD = '\033[31m'
            COLORS_NEW = '\033[32m'
            COLORS_COMMIT = '\033[33m'
            COLORS_FILES = '\033[34m'
            COLORS_ENDC = '\033[0m'
        prefix_old = "-"
        prefix_new = "+"
        format_dict = {'old_value':
                       {'color': COLORS_OLD, 'prefix': prefix_old},
                       'new_value':
                       {'color': COLORS_NEW, 'prefix': prefix_new}}
        if oldfile or newfile:
            print(f"kas diff {oldfile} {newfile}")
            print(prefix_old * 3, f" {oldfile}")
            print(prefix_new * 3, f" {newfile}")
        vc_dict = diff_output.get('values_changed', {})
        vcs_dict = diff_output.get('vcs', {})
        if vc_dict and not commit_only:
            print(f"{COLORS_FILES}@@ config changed @@{COLORS_ENDC}")
            for key in vc_dict.keys():
                oldval = vc_dict[key]['old_value']
                newval = vc_dict[key]['new_value']
                if isinstance(oldval, str) and isinstance(newval, str):
                    print_unified_diff(oldval, newval, key)
                    continue
                for k, i in format_dict.items():
                    value = vc_dict[key][k]
                    if not value:
                        continue
                    value = str(value).replace('\n', f"\n{i['prefix']}")
                    print(f"{i['color']}"
                          f"{i['prefix']}{' ' * 4}{key}: {value}"
                          f"{COLORS_ENDC}")
        if vcs_dict and not content_only:
            for key in vcs_dict.keys():
                print(f"{COLORS_FILES}@@ {key} commits diff @@{COLORS_ENDC}")
                for li in vcs_dict[key]:
                    msg = li['message']
                    if oneline:
                        msg = msg.split('\n')[0]
                        print(f"{COLORS_COMMIT}"
                              f"{li['commit'][:7]}{COLORS_ENDC}"
                              f" {msg}")
                    else:
                        print(f"{COLORS_COMMIT}{li['commit']}: {COLORS_ENDC}"
                              f"{li['author']} {li['commit_date']}")
                        indented_msg = '\n'.join([' ' * 4 + s
                                                 for s in msg.split('\n')])
                        print(indented_msg)
                if key != list(vcs_dict.keys())[-1]:
                    print("---")

    def run(self, args):
        args.skip += [
            'setup_environ',
            'write_bbconfig'
        ]
        ctx = create_global_context(args)
        ctx.config = Config(ctx, args.config1)
        macro = Macro()
        macro.run(ctx, args.skip)
        config1 = ctx.config.get_config(remove_includes=True,
                                        apply_overrides=True)
        repos1 = ctx.config.get_repos()

        args.skip += [
            'setup_dir',
            'setup_home',
            'setup_ssh_agent'
        ]
        ctx.config = Config(ctx, args.config2)
        macro.run(ctx, args.skip)
        config2 = ctx.config.get_config(remove_includes=True,
                                        apply_overrides=True)
        repos2 = ctx.config.get_repos()

        diff = Diff.compare_dicts(config1, config2)
        diff_output = {}
        if diff:
            # combine the diff output and the repo diff to a dict
            diff_output.update(diff)

            vcs_dict = {}
            # check for commit IDs/tags of repos and compare them
            for key in diff.get('values_changed', {}):
                if 'commit' in key or 'tag' in key:
                    commit1 = diff['values_changed'][key]['old_value']
                    commit2 = diff['values_changed'][key]['new_value']
                    repo_name_arr = key.split('.')
                    if len(repo_name_arr) < 3 or repo_name_arr[0] != 'repos':
                        continue
                    repo_name = repo_name_arr[1]
                    # we do not known which config contains the latest commit,
                    # so we need to check both configs
                    for each_repo in repos1:
                        if each_repo.path and each_repo.url \
                           and each_repo.name == repo_name:
                            repo_diff = each_repo.diff(commit2, commit1)
                            if len(repo_diff.get(each_repo.name)) > 0:
                                vcs_dict.update(repo_diff)

                    for each_repo in repos2:
                        if each_repo.path and each_repo.url \
                           and each_repo.name == repo_name:
                            repo_diff = each_repo.diff(commit1, commit2)
                            if len(repo_diff.get(each_repo.name)) > 0:
                                vcs_dict.update(repo_diff)
        if vcs_dict:
            diff_output['vcs'] = vcs_dict
        if args.format == 'json':
            print(json.dumps(diff_output, indent=4))
        else:
            Diff.formatting_diff_output(args.config1, args.config2,
                                        diff_output, args.oneline,
                                        args.no_color, args.commit_only,
                                        args.content_only)


__KAS_PLUGINS__ = [Diff]



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens, 2025
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This plugin implements the ``kas {clean,cleansstate,cleanall}``
    commands. In case a configuration file is provided, it will be used to
    determine the build system and the files managed by kas.
"""

import os
import shutil
import logging
import subprocess
from pathlib import Path
from kas.context import create_global_context, get_context
from kas.config import Config, CONFIG_YAML_FILE
from kas.includehandler import ConfigFile
from kas.libcmds import Macro
from kas.kasusererror import KasUserError
from kas import plugins


__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens, 2025'


class Clean():
    """
    Clean the build artifacts by removing the build artifacts
    directory.
    """

    name = 'clean'
    helpmsg = (
        'Clean build artifacts, keep sstate cache and downloads.'
    )
    config_files = None

    @classmethod
    def setup_parser(cls, parser):
        parser.add_argument('--dry-run',
                            action='store_true',
                            default=False,
                            help='Do not remove anything, just print what '
                                 'would be removed')
        parser.add_argument('--isar',
                            action='store_true',
                            default=False,
                            help='Use ISAR build directory layout')
        parser.add_argument('config',
                            help='Config file(s), separated by colon. Using '
                                 '.config.yaml in KAS_WORK_DIR if existing '
                                 'and none is specified.',
                            nargs='?')

    def run(self, args):
        ctx = create_global_context(args)
        default_conf_file = Path(ctx.kas_work_dir) / CONFIG_YAML_FILE
        build_system = None
        if args.config:
            self.config_files = args.config
        elif default_conf_file.exists():
            self.config_files = str(default_conf_file)
        if self.config_files:
            # By definition, build_system key must be present in the first
            # config file to take effect.
            cf = ConfigFile.load(self.config_files.split(':')[0], False, False)
            build_system = cf.config.get('build_system')
        if args.isar:
            build_system = 'isar'

        logging.debug('Run clean in "%s" mode' % (build_system or 'default'))
        if args.dry_run:
            logging.warning('Dry run, not removing anything')
        tmpdirs = Path(ctx.build_dir).glob('tmp*')

        dirs_to_remove = []
        for tmpdir in tmpdirs:
            logging.info(f'Removing {tmpdir}')
            if build_system == 'isar':
                dirs_to_remove.append(tmpdir)
            else:
                if not args.dry_run:
                    shutil.rmtree(tmpdir)

        if len(dirs_to_remove) == 0:
            return

        clean_args = ['sudo', '--prompt', '[sudo] enter password for %U '
                      'to clean ISAR artifacts']
        clean_args.extend(['rm', '-rf'])
        clean_args.extend([p.as_posix() for p in dirs_to_remove])
        logging.debug(' '.join(clean_args))
        if not args.dry_run:
            subprocess.check_call(clean_args)

    @staticmethod
    def clear_dir_content(directory):
        """
        Clear the contents of a directory without removing the dir itself.
        """
        for item in directory.iterdir():
            if item.is_dir():
                shutil.rmtree(item)
            else:
                item.unlink()


class CleanSstate(Clean):
    """
    Removes the build artifacts and the empties the sstate cache.
    """

    name = 'cleansstate'
    helpmsg = (
        'Clean build artifacts and sstate cache.'
    )

    def run(self, args):
        super().run(args)
        ctx = get_context()
        sstate_dir = Path(os.environ.get('SSTATE_DIR',
                                         Path(ctx.build_dir) / 'sstate-cache'))
        if sstate_dir.exists():
            logging.info(f'Removing {sstate_dir}/*')
            if not args.dry_run:
                self.clear_dir_content(sstate_dir)


class CleanAll(CleanSstate):
    """
    Removes the build artifacts, empties the sstate cache and the downloads.
    """

    name = 'cleanall'
    helpmsg = (
        'Clean build artifacts, sstate-cache and downloads.'
    )

    def run(self, args):
        super().run(args)
        ctx = get_context()
        downloads_dir = Path(os.environ.get('DL_DIR',
                                            Path(ctx.build_dir) / 'downloads'))
        if downloads_dir.exists():
            logging.info(f'Removing {downloads_dir}/*')
            if not args.dry_run:
                self.clear_dir_content(downloads_dir)


class Purge(CleanAll):
    """
    Clears the contents of the build directory, sstate-cache, downloads and
    the repos managed by kas (including referenced repos in
    ``KAS_REPO_REF_DIR``, if set). In ``KAS_WORK_DIR`` it will remove the
    default configuration file and the ``KAS_BUILD_DIR`` (if present).
    To preserve the reference repositories, run with ``--preserve-repo-refs``.
    This command requires a configuration file to locate the managed repos.

    .. note::
        Before purging, kas needs to checkout and resolve all repos to locate
        the repos managed by kas.
    """

    name = 'purge'
    helpmsg = (
        'Purge all data managed by kas, including managed repos.'
    )

    @classmethod
    def setup_parser(cls, parser):
        super().setup_parser(parser)
        parser.add_argument('--preserve-repo-refs',
                            action='store_true',
                            default=False,
                            help='Do not remove the reference repositories')

    def run(self, args):
        super().run(args)
        ctx = get_context()

        if not self.config_files:
            raise KasUserError('Purge requires a config file to locate '
                               'managed repos.')

        ctx.config = Config(ctx, self.config_files)
        # to read the config, we need all repos (but no build env),
        macro = Macro()
        macro.run(ctx, skip=['repos_apply_patches', 'write_bb_config',
                             'setup_environ'])

        for r in ctx.config.get_repos():
            if r.operations_disabled:
                logging.debug(f'Skipping {r.name} as not managed by kas')
                continue
            logging.info(f'Removing {r.path}')
            if not args.dry_run:
                shutil.rmtree(r.path)
            if ctx.kas_repo_ref_dir and not args.preserve_repo_refs:
                ref_repo = Path(ctx.kas_repo_ref_dir) / r.qualified_name
                if ref_repo.exists():
                    logging.info(f'Removing {ref_repo}')
                    if not args.dry_run:
                        shutil.rmtree(ref_repo)

        build_dir = Path(ctx.build_dir)
        logging.info(f'Removing {build_dir}/*')
        if not args.dry_run:
            self.clear_dir_content(build_dir)

        if os.environ.get("KAS_BUILDTOOLS_DIR"):
            buildtools_dir = (
                Path(os.environ.get("KAS_BUILDTOOLS_DIR")).resolve()
            )
            if buildtools_dir.exists():
                logging.info(f'Removing {buildtools_dir}')
                if not args.dry_run:
                    shutil.rmtree(buildtools_dir)

        work_dir = Path(ctx.kas_work_dir)
        default_config = work_dir / CONFIG_YAML_FILE
        if default_config.exists():
            logging.info(f'Removing {default_config}')
            if not args.dry_run:
                default_config.unlink()

        clean_paths = list(ctx.managed_paths)
        # Plugins can register additional paths by providing get_managed_paths.
        # These paths must be relative to the work_dir.
        for plugin in plugins.all():
            if hasattr(plugin, 'get_managed_paths'):
                ppaths = plugin.get_managed_paths()
                clean_paths.extend([work_dir / p for p in ppaths])

        for path in [Path(p) for p in clean_paths if Path(p).exists()]:
            logging.info(f'Removing {path}')
            if not args.dry_run:
                if path.is_file() or path.is_symlink():
                    path.unlink()
                else:
                    shutil.rmtree(path)


__KAS_PLUGINS__ = [Clean, CleanSstate, CleanAll, Purge]
